/* Custom Symbols */
code.doc-symbol-parameter::after {
    content: "param";
}

code.doc-symbol-attribute::after {
    content: "attr";
}

code.doc-symbol-function::after {
    content: "def";
}

code.doc-symbol-method::after {
    content: "def";
}

code.doc-symbol-class::after {
    content: "class";
}

code.doc-symbol-module::after {
    content: "module";
}

/* Additional symbols for YAML models */
code.doc-symbol-model {
    color: var(--doc-symbol-class-fg-color);
    background-color: var(--doc-symbol-class-bg-color);
}

code.doc-symbol-model::after {
    content: "model";
}

code.doc-symbol-abstract-model {
    color: var(--doc-symbol-method-fg-color);
    background-color: var(--doc-symbol-method-bg-color);
}

code.doc-symbol-abstract-model::after {
    content: "*model";
}
code.doc-symbol-field {
    color: var(--doc-symbol-attribute-fg-color);
    background-color: var(--doc-symbol-attribute-bg-color);
}

code.doc-symbol-field::after {
    content: "field";
}

/* Abbreviate module symbol on full-width navigation */
@media screen and (min-width: 76.25em) {
    .md-nav code.doc-symbol-nav.doc-symbol-module::after {
        content: "M";
    }
}

/* Indentation. */
div.doc-contents:not(.first) {
    padding-left: 25px;
    border-left: .05rem solid var(--md-typeset-table-color);
}

/* Allow module name to wrap */
doc-module-name {
    overflow-wrap: break-word;
}

/* Mark external links as such. */
a.external::after,
a.autorefs-external::after {
    /* https://primer.style/octicons/arrow-up-right-24 */
    mask-image: url('data:image/svg+xml,<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24"><path d="M18.25 15.5a.75.75 0 00.75-.75v-9a.75.75 0 00-.75-.75h-9a.75.75 0 000 1.5h7.19L6.22 16.72a.75.75 0 101.06 1.06L17.5 7.56v7.19c0 .414.336.75.75.75z"></path></svg>');
    -webkit-mask-image: url('data:image/svg+xml,<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 24 24"><path d="M18.25 15.5a.75.75 0 00.75-.75v-9a.75.75 0 00-.75-.75h-9a.75.75 0 000 1.5h7.19L6.22 16.72a.75.75 0 101.06 1.06L17.5 7.56v7.19c0 .414.336.75.75.75z"></path></svg>');
    content: ' ';

    display: inline-block;
    vertical-align: middle;
    position: relative;

    height: 1em;
    width: 1em;
    background-color: currentColor;
}

a.external:hover::after,
a.autorefs-external:hover::after {
    background-color: var(--md-accent-fg-color);
}
